import logging
import os


def get_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)
    try:
        from uvicorn.logging import DefaultFormatter

        formatter = DefaultFormatter(fmt="%(levelprefix)s %(name)s %(message)s")
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    except ImportError:
        logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    logger.setLevel(os.environ.get("LOG_LEVEL", "INFO"))
    return logger


logger = get_logger("rsserpent")



from typing import Any

from pyquery import PyQuery

from ...utils import Browser

path = "/_/example/playwright"


async def provider() -> dict[str, Any]:
    """Define a basic example data provider function."""
    async with Browser() as browser:
        await browser.goto("https://httpbin.org/html")
        content = await browser.content()
    dom = PyQuery(content)
    return {
        "title": "Example",
        "link": "https://example.com",
        "description": "An example rsserpent plugin.",
        "items": [{"title": dom("h1").text(), "description": dom("div > p").text()}],
    }



from typing import Any

path = "/_/example"


async def provider() -> dict[str, Any]:
    """Define a basic example data provider function."""
    return {
        "title": "Example",
        "link": "https://example.com",
        "description": "An example rsserpent plugin.",
        "items": [{"title": "Example Title", "description": "Example Description", "link": "https://example.com"}],
    }



from feedgen.feed import FeedGenerator

from rsserpent_rev.models.plugin import Feed

path = "/_/example/feedgen"


async def provider() -> Feed:
    """Define a basic example data provider function."""
    fg = FeedGenerator()
    fg.id("http://lernfunk.de/media/654321")
    fg.title("Some Testfeed")
    fg.author({"name": "John Doe", "email": "john@example.de"})
    fg.link(href="http://example.com", rel="alternate")
    fg.logo("http://ex.com/logo.jpg")
    fg.subtitle("This is a cool feed!")
    fg.link(href="http://larskiesow.de/test.atom", rel="self")
    fg.language("en")
    fe = fg.add_entry()
    fe.id("http://lernfunk.de/media/654322")
    fe.title("A second test")
    fe.link(href="http://example.com/2", rel="alternate")
    fe.content("And here the content")
    return fg



import copy
import hashlib
import os
import re
import subprocess
import sys
import traceback
from pathlib import Path

import arrow
import feedgen
import feedgen.feed
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route
from starlette.templating import Jinja2Templates
from starlette.templating import _TemplateResponse as TemplateResponse

from .log import logger
from .models import Feed, Item, Plugin, ProviderFn
from .plugins import plugins
from .utils import fetch_data

templates = Jinja2Templates(directory=str(Path(__file__).parent / "templates"))
templates.env.autoescape = True
templates.env.globals["arrow"] = arrow


async def exception_handler(request: Request, exception: Exception) -> TemplateResponse:
    """Return an HTML web page for displaying the current exception."""
    tb = traceback.format_exc()
    for path in sys.path:
        tb, _ = re.subn(rf'(?<=File "){re.escape(path)}[/\\]', "", tb)
    return templates.TemplateResponse(
        "exception.html.jinja",
        {
            "exception": exception,
            "request": request,
            "traceback": tb,
        },
    )


def startup() -> None:
    """Install Chromium on start-up."""
    if sys.platform == "linux" and "PLAYWRIGHT_BROWSERS_PATH" not in os.environ:
        os.environ["PLAYWRIGHT_BROWSERS_PATH"] = "/tmp"  # pragma: is_not_linux
    logger.info(f"Running on {sys.platform}")
    logger.info(f"Installing Chromium at {os.environ.get('PLAYWRIGHT_BROWSERS_PATH')}")
    subprocess.run("playwright install chromium".split())


async def index(request: Request) -> TemplateResponse:
    """Return the index HTML web page."""
    return templates.TemplateResponse("index.html.jinja", {"request": request})


routes = [Route("/", endpoint=index)]


def filter_fg(fg: feedgen.feed.FeedGenerator, request: Request) -> None:
    p = request.query_params
    new_entry = [
        entry
        for entry in fg.entry()
        if ("title_include" not in p or re.search(p["title_include"], entry.title()))
        and ("title_exclude" not in p or not re.search(p["title_exclude"], entry.title()))
        and ("description_include" not in p or re.search(p["description_include"], entry.description()))
        and ("description_exclude" not in p or not re.search(p["description_exclude"], entry.description()))
    ]
    if "limit" in p:
        new_entry = new_entry[: int(p["limit"])]
    fg.entry(new_entry, replace=True)


def gen_ids(fg: feedgen.feed.FeedGenerator) -> None:
    if not fg.id():
        fg.id(hashlib.md5(fg.title().encode()).hexdigest())
    for entry in fg.entry():
        if not entry.id():
            hash_content = entry.title() + entry.description()
            entry.id(hashlib.md5(hash_content.encode()).hexdigest())


for plugin in plugins:
    for path, provider in plugin.routers.items():

        async def endpoint(request: Request, plugin: Plugin = plugin, provider: ProviderFn = provider) -> Response:
            """Return an RSS feed of XML format."""
            path_params = request.path_params
            for key, value in request.path_params.items():
                if isinstance(value, str) and (value.endswith(".rss") or value.endswith(".atom")):
                    path_params[key] = value[:-5]
            data = await fetch_data(provider, request.path_params, dict(request.query_params))
            if isinstance(data, dict):
                rss20_feed = Feed(**data)
                if "items" in data:
                    rss20_feed.items = [Item(**item) for item in data["items"]]
                rss20_feed.apply_defaults(plugin)
                fg = copy.copy(rss20_feed.to_feedgen())
                filter_fg(fg, request)
                if request.url.path.endswith(".atom"):
                    gen_ids(fg)
                    return Response(content=fg.atom_str(pretty=True), media_type="application/atom+xml")
                return Response(content=fg.rss_str(pretty=True), media_type="application/xml")
            else:
                fg = copy.copy(data)
                filter_fg(fg, request)
                if request.url.path.endswith(".rss"):
                    return Response(content=fg.rss_str(pretty=True), media_type="application/xml")
                return Response(content=fg.atom_str(pretty=True), media_type="application/atom+xml")

        routes.append(Route(path, endpoint=endpoint))
        routes.append(Route(path + ".rss", endpoint=endpoint))
        routes.append(Route(path + ".atom", endpoint=endpoint))


async def all_routes(request: Request) -> TemplateResponse:
    """Return all available routes."""
    return templates.TemplateResponse(
        "all_routes.html.jinja",
        {"plugins": plugins, "request": request},
    )


routes.append(Route("/all_routes", endpoint=all_routes))

app = Starlette(
    debug=bool(os.environ.get("DEBUG")),
    exception_handlers={Exception: exception_handler},
    on_startup=[startup],
    routes=routes,
)



from typing import TYPE_CHECKING, Any

import arrow
from arrow import Arrow
from feedgen.feed import FeedGenerator
from pydantic import BaseModel, Field, root_validator

from . import Plugin

if TYPE_CHECKING:
    AnyUrl = str
    HttpUrl = str
else:
    from pydantic import AnyUrl, HttpUrl


class RSSModelError(ValueError):
    """Exception for `Feed` model validation error."""

    empty_title_and_description = "at least one of <title> or <description> must be present."


class Category(BaseModel):
    """Data model for the `<category>` field in an RSS 2.0 feed."""

    name: str
    domain: HttpUrl | None


class Enclosure(BaseModel):
    """Data model for the `<enclosure>` field in an RSS 2.0 feed."""

    length: int = 0
    type: str
    url: AnyUrl


class Guid(BaseModel):
    """Data model for the `<guid>` field in an RSS 2.0 feed."""

    is_perma_link: bool = True
    value: str


class Image(BaseModel):
    """Data model for the `<image>` field in an RSS 2.0 feed."""

    url: HttpUrl
    title: str
    link: HttpUrl
    width: int | None = 88
    height: int | None = 31
    description: str | None


class Source(BaseModel):
    """Data model for the `<source>` field in an RSS 2.0 feed."""

    name: str
    url: HttpUrl | None


class Item(BaseModel):
    """Data model for the `<item>` field in an RSS 2.0 feed."""

    title: str | None
    link: HttpUrl | None
    description: str | None
    author: str | None
    categories: list[Category] | None
    comments: HttpUrl | None
    enclosure: Enclosure | None
    guid: Guid | None
    pub_date: Arrow | None
    source: Source | None

    class Config:  # noqa: D106
        arbitrary_types_allowed = True

    @root_validator
    def validate(  # type: ignore[override]
        cls,  # noqa: N805
        values: dict[str, Any],  # noqa: N805
    ) -> dict[str, Any]:
        """Ensure at least one of `<title>` or `<description>` is present."""
        title, description = values.get("title"), values.get("description")
        if title is None and description is None:
            raise RSSModelError(RSSModelError.empty_title_and_description)
        return values


class Feed(BaseModel):
    """Data model for RSS 2.0 feeds.

    See specification at https://www.rssboard.org/rss-specification.
    Note that some rarely used fields are not implemented:
    * `<cloud>`
    * `<rating>`
    * `<textInput>`
    * `<skipHours>`
    * `<skipDays>`
    """

    title: str
    link: HttpUrl
    description: str
    language: str | None
    copyright: str | None
    managing_editor: str | None
    web_master: str | None
    pub_date: Arrow | None = Field(default_factory=arrow.utcnow)
    last_build_date: Arrow | None = Field(default_factory=arrow.utcnow)
    categories: list[Category] | None
    generator: str | None = __package__.split(".")[0]
    docs: HttpUrl | None = "https://www.rssboard.org/rss-specification"
    ttl: int | None = 60
    image: Image | None
    items: list[Item] | None

    class Config:  # noqa: D106
        arbitrary_types_allowed = True

    def apply_defaults(self, plugin: Plugin) -> None:
        self.managing_editor = (
            f"{plugin.author.email} ({plugin.author.name})" if self.managing_editor is None else self.managing_editor
        )
        self.web_master = (
            f"{plugin.author.email} ({plugin.author.name})" if self.web_master is None else self.web_master
        )
        self.last_build_date = arrow.utcnow() if self.last_build_date is None else self.last_build_date
        self.pub_date = arrow.utcnow() if self.pub_date is None else self.pub_date

    def to_feedgen(self) -> FeedGenerator:
        """Convert the `Feed` model to a `feedgen.feed.FeedGenerator` object."""
        fg = FeedGenerator()
        fg.title(self.title)
        fg.link(href=self.link)
        fg.description(self.description)
        fg.language(self.language)
        fg.copyright(self.copyright)
        fg.managingEditor(self.managing_editor)
        fg.webMaster(self.web_master)
        fg.lastBuildDate(self.last_build_date.datetime)
        fg.generator(self.generator)
        fg.docs(self.docs)
        fg.ttl(self.ttl)
        if self.image:
            fg.image(
                url=self.image.url,
                title=self.image.title,
                link=self.image.link,
                width=self.image.width,
                height=self.image.height,
                description=self.image.description,
            )
        if self.categories:
            for category in self.categories:
                fg.category(category.name, domain=category.domain)
        if self.items:
            for item in self.items:
                fe = fg.add_entry()
                fe.title(item.title)
                fe.link(href=item.link)
                fe.description(item.description)
                fe.author(item.author)
                fe.comments(item.comments)
                if item.enclosure:
                    fe.enclosure(
                        url=item.enclosure.url,
                        length=item.enclosure.length,
                        type=item.enclosure.type,
                    )
                if item.guid:
                    fe.guid(item.guid.value, isPermaLink=item.guid.is_perma_link)
                if item.pub_date:
                    fe.pubDate(item.pub_date.datetime)
                else:
                    fe.pubDate(arrow.utcnow().datetime)
                if item.source:
                    fe.source(title=item.source.name, url=item.source.url)
                if item.categories:
                    for category in item.categories:
                        fe.category(category.name, domain=category.domain)
        return fg



from typing import Any

from rsserpent_rev.log import get_logger

path = "/_/example/log"

logger = get_logger(__name__)


async def provider() -> dict[str, Any]:
    """Define a basic example data provider function."""
    logger.info("Example plugin is called.")
    return {
        "title": "Example",
        "link": "https://example.com",
        "description": "An example rsserpent plugin.",
        "items": [{"title": "Example Title", "description": "Example Description"}],
    }



from typing import Any

path = "/_/example/{n:int}"


async def provider(n: int) -> dict[str, Any]:
    # """Define a example data provider function with arguments."""
    """Define a example data provider function with arguments.

    Args:
        n (int): The number of items to generate.

    Returns:
        Dict[str, Any]: A dictionary containing the feed data.
    """

    return {
        "title": "Example",
        "link": "https://example.com",
        "description": "An example rsserpent plugin.",
        "items": [{"title": f"Example Title {i}", "description": f"Example Description {i}"} for i in range(1, n + 1)],
    }
